@@ -9,9 +9,9 @@
 #include "nvim/api/private/defs.h"
 #include "nvim/api/private/helpers.h"
 #include "nvim/ascii.h"
+#include "nvim/autocmd.h"
 #include "nvim/buffer.h"
 #include "nvim/eval/typval.h"
-#include "nvim/fileio.h"
 #include "nvim/lua/executor.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
@@ -23,15 +23,14 @@
 // Copy string or array of strings into an empty array.
 // Get the event number, unless it is an error. Then goto `goto_name`.
 #define GET_ONE_EVENT(event_nr, event_str, goto_name) \
-  char_u *__next_ev; \
+  char *__next_ev; \
   event_T event_nr = \
-    event_name2nr((char_u *)event_str.data.string.data, &__next_ev); \
+    event_name2nr(event_str.data.string.data, &__next_ev); \
   if (event_nr >= NUM_EVENTS) { \
     api_set_error(err, kErrorTypeValidation, "unexpected event"); \
     goto goto_name; \
   }
 
-
 // ID for associating autocmds created via nvim_create_autocmd
 // Used to delete autocmds from nvim_del_autocmd
 static int64_t next_autocmd_id = 1;
@@ -60,6 +59,9 @@ static int64_t next_autocmd_id = 1;
 ///             - group (string|integer): the autocommand group name or id to match against.
 ///             - event (string|array): event or events to match against |autocmd-events|.
 ///             - pattern (string|array): pattern or patterns to match against |autocmd-pattern|.
+///             Cannot be used with {buffer}
+///             - buffer: Buffer number or list of buffer numbers for buffer local autocommands
+///             |autocmd-buflocal|. Cannot be used with {pattern}
 /// @return Array of autocommands matching the criteria, with each item
 ///         containing the following fields:
 ///             - id (number): the autocommand id (only when defined with the API).
@@ -67,7 +69,9 @@ static int64_t next_autocmd_id = 1;
 ///             - group_name (string): the autocommand group name.
 ///             - desc (string): the autocommand description.
 ///             - event (string): the autocommand event.
-///             - command (string): the autocommand command.
+///             - command (string): the autocommand command. Note: this will be empty if a callback is set.
+///             - callback (function|string|nil): Lua function or name of a Vim script function
+///             which is executed when this autocommand is triggered.
 ///             - once (boolean): whether the autocommand is only run once.
 ///             - pattern (string): the autocommand pattern.
 ///             If the autocommand is buffer local |autocmd-buffer-local|:
@@ -79,8 +83,8 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
   // TODO(tjdevries): Would be cool to add nvim_get_autocmds({ id = ... })
 
   Array autocmd_list = ARRAY_DICT_INIT;
-  char_u *pattern_filters[AUCMD_MAX_PATTERNS];
-  char_u pattern_buflocal[BUFLOCAL_PAT_LEN];
+  char *pattern_filters[AUCMD_MAX_PATTERNS];
+  char pattern_buflocal[BUFLOCAL_PAT_LEN];
 
   Array buffers = ARRAY_DICT_INIT;
 
@@ -149,7 +153,7 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
   if (opts->pattern.type != kObjectTypeNil) {
     Object v = opts->pattern;
     if (v.type == kObjectTypeString) {
-      pattern_filters[pattern_filter_count] = (char_u *)v.data.string.data;
+      pattern_filters[pattern_filter_count] = v.data.string.data;
       pattern_filter_count += 1;
     } else if (v.type == kObjectTypeArray) {
       if (v.data.array.size > AUCMD_MAX_PATTERNS) {
@@ -165,7 +169,7 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
           goto cleanup;
         }
 
-        pattern_filters[pattern_filter_count] = (char_u *)item.data.string.data;
+        pattern_filters[pattern_filter_count] = item.data.string.data;
         pattern_filter_count += 1;
       });
     } else {
@@ -212,7 +216,7 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
   }
 
   FOREACH_ITEM(buffers, bufnr, {
-    pattern_filters[pattern_filter_count] = (char_u *)bufnr.data.string.data;
+    pattern_filters[pattern_filter_count] = bufnr.data.string.data;
     pattern_filter_count += 1;
   });
 
@@ -238,8 +242,8 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
           assert(i < AUCMD_MAX_PATTERNS);
           assert(pattern_filters[i]);
 
-          char_u *pat = pattern_filters[i];
-          int patlen = (int)STRLEN(pat);
+          char *pat = pattern_filters[i];
+          int patlen = (int)strlen(pat);
 
           if (aupat_is_buflocal(pat, patlen)) {
             aupat_normalize_buflocal_pat(pattern_buflocal,
@@ -250,7 +254,7 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
             pat = pattern_buflocal;
           }
 
-          if (strequal((char *)ap->pat, (char *)pat)) {
+          if (strequal(ap->pat, pat)) {
             passed = true;
             break;
           }
@@ -281,9 +285,28 @@ Array nvim_get_autocmds(Dict(get_autocmds) *opts, Error *err)
           PUT(autocmd_info, "desc", CSTR_TO_OBJ(ac->desc));
         }
 
-        PUT(autocmd_info,
-            "command",
-            STRING_OBJ(cstr_as_string(aucmd_exec_to_string(ac, ac->exec))));
+        if (ac->exec.type == CALLABLE_CB) {
+          PUT(autocmd_info, "command", STRING_OBJ(STRING_INIT));
+
+          Callback *cb = &ac->exec.callable.cb;
+          switch (cb->type) {
+          case kCallbackLua:
+            if (nlua_ref_is_function(cb->data.luaref)) {
+              PUT(autocmd_info, "callback", LUAREF_OBJ(api_new_luaref(cb->data.luaref)));
+            }
+            break;
+          case kCallbackFuncref:
+          case kCallbackPartial:
+            PUT(autocmd_info, "callback", STRING_OBJ(cstr_as_string(callback_to_string(cb))));
+            break;
+          default:
+            abort();
+          }
+        } else {
+          PUT(autocmd_info,
+              "command",
+              STRING_OBJ(cstr_as_string(xstrdup(ac->exec.callable.cmd))));
+        }
 
         PUT(autocmd_info,
             "pattern",
@@ -348,6 +371,22 @@ cleanup:
 ///     })
 /// </pre>
 ///
+/// Lua functions receive a table with information about the autocmd event as an argument. To use
+/// a function which itself accepts another (optional) parameter, wrap the function
+/// in a lambda:
+///
+/// <pre>
+///     -- Lua function with an optional parameter.
+///     -- The autocmd callback would pass a table as argument but this
+///     -- function expects number|nil
+///     local myluafun = function(bufnr) bufnr = bufnr or vim.api.nvim_get_current_buf() end
+///
+///     vim.api.nvim_create_autocmd({"BufEnter", "BufWinEnter"}, {
+///       pattern = {"*.c", "*.h"},
+///       callback = function() myluafun() end,
+///     })
+/// </pre>
+///
 /// Example using command:
 /// <pre>
 ///     vim.api.nvim_create_autocmd({"BufEnter", "BufWinEnter"}, {
@@ -389,6 +428,7 @@ cleanup:
 ///                 - match: (string) the expanded value of |<amatch>|
 ///                 - buf: (number) the expanded value of |<abuf>|
 ///                 - file: (string) the expanded value of |<afile>|
+///                 - data: (any) arbitrary data passed to |nvim_exec_autocmds()|
 ///             - command (string) optional: Vim command to execute on event. Cannot be used with
 ///             {callback}
 ///             - once (boolean) optional: defaults to false. Run the autocommand
@@ -412,7 +452,6 @@ Integer nvim_create_autocmd(uint64_t channel_id, Object event, Dict(create_autoc
   AucmdExecutable aucmd = AUCMD_EXECUTABLE_INIT;
   Callback cb = CALLBACK_NONE;
 
-
   if (!unpack_string_or_array(&event_array, &event, "event", true, err)) {
     goto cleanup;
   }
@@ -427,7 +466,8 @@ Integer nvim_create_autocmd(uint64_t channel_id, Object event, Dict(create_autoc
     // not do that.
 
     Object *callback = &opts->callback;
-    if (callback->type == kObjectTypeLuaRef) {
+    switch (callback->type) {
+    case kObjectTypeLuaRef:
       if (callback->data.luaref == LUA_NOREF) {
         api_set_error(err,
                       kErrorTypeValidation,
@@ -444,10 +484,12 @@ Integer nvim_create_autocmd(uint64_t channel_id, Object event, Dict(create_autoc
 
       cb.type = kCallbackLua;
       cb.data.luaref = api_new_luaref(callback->data.luaref);
-    } else if (callback->type == kObjectTypeString) {
+      break;
+    case kObjectTypeString:
       cb.type = kCallbackFuncref;
-      cb.data.funcref = vim_strsave((char_u *)callback->data.string.data);
-    } else {
+      cb.data.funcref = string_to_cstr(callback->data.string);
+      break;
+    default:
       api_set_error(err,
                     kErrorTypeException,
                     "'callback' must be a lua function or name of vim function");
@@ -460,7 +502,7 @@ Integer nvim_create_autocmd(uint64_t channel_id, Object event, Dict(create_autoc
     Object *command = &opts->command;
     if (command->type == kObjectTypeString) {
       aucmd.type = CALLABLE_EX;
-      aucmd.callable.cmd = vim_strsave((char_u *)command->data.string.data);
+      aucmd.callable.cmd = string_to_cstr(command->data.string);
     } else {
       api_set_error(err,
                     kErrorTypeValidation,
@@ -510,14 +552,12 @@ Integer nvim_create_autocmd(uint64_t channel_id, Object event, Dict(create_autoc
 
     int retval;
 
-    for (size_t i = 0; i < patterns.size; i++) {
-      Object pat = patterns.items[i];
-
+    FOREACH_ITEM(patterns, pat, {
       // See: TODO(sctx)
       WITH_SCRIPT_CONTEXT(channel_id, {
         retval = autocmd_register(autocmd_id,
                                   event_nr,
-                                  (char_u *)pat.data.string.data,
+                                  pat.data.string.data,
                                   (int)pat.data.string.size,
                                   au_group,
                                   is_once,
@@ -530,10 +570,9 @@ Integer nvim_create_autocmd(uint64_t channel_id, Object event, Dict(create_autoc
         api_set_error(err, kErrorTypeException, "Failed to set autocmd");
         goto cleanup;
       }
-    }
+    })
   });
 
-
 cleanup:
   aucmd_exec_free(&aucmd);
   api_free_array(event_array);
@@ -560,7 +599,7 @@ void nvim_del_autocmd(Integer id, Error *err)
 }
 
 /// Clear all autocommands that match the corresponding {opts}. To delete
-/// a particular autocmd, see |nvim_del_autocmd|.
+/// a particular autocmd, see |nvim_del_autocmd()|.
 /// @param opts Parameters
 ///         - event: (string|table)
 ///              Examples:
@@ -621,8 +660,8 @@ void nvim_clear_autocmds(Dict(clear_autocmds) *opts, Error *err)
   if (event_array.size == 0) {
     FOR_ALL_AUEVENTS(event) {
       FOREACH_ITEM(patterns, pat_object, {
-        char_u *pat = (char_u *)pat_object.data.string.data;
-        if (!clear_autocmd(event, pat, au_group, err)) {
+        char *pat = pat_object.data.string.data;
+        if (!clear_autocmd(event, (char *)pat, au_group, err)) {
           goto cleanup;
         }
       });
@@ -632,8 +671,8 @@ void nvim_clear_autocmds(Dict(clear_autocmds) *opts, Error *err)
       GET_ONE_EVENT(event_nr, event_str, cleanup);
 
       FOREACH_ITEM(patterns, pat_object, {
-        char_u *pat = (char_u *)pat_object.data.string.data;
-        if (!clear_autocmd(event_nr, pat, au_group, err)) {
+        char *pat = pat_object.data.string.data;
+        if (!clear_autocmd(event_nr, (char *)pat, au_group, err)) {
           goto cleanup;
         }
       });
@@ -643,8 +682,6 @@ void nvim_clear_autocmds(Dict(clear_autocmds) *opts, Error *err)
 cleanup:
   api_free_array(event_array);
   api_free_array(patterns);
-
-  return;
 }
 
 /// Create or get an autocommand group |autocmd-groups|.
@@ -691,7 +728,7 @@ Integer nvim_create_augroup(uint64_t channel_id, String name, Dict(create_augrou
 ///
 /// To get a group id one can use |nvim_get_autocmds()|.
 ///
-/// NOTE: behavior differs from |augroup-delete|. When deleting a group, autocommands contained in
+/// NOTE: behavior differs from |:augroup-delete|. When deleting a group, autocommands contained in
 /// this group will also be deleted and cleared. This group will no longer exist.
 /// @param id Integer The id of the group.
 /// @see |nvim_del_augroup_by_name()|
@@ -709,10 +746,10 @@ void nvim_del_augroup_by_id(Integer id, Error *err)
 
 /// Delete an autocommand group by name.
 ///
-/// NOTE: behavior differs from |augroup-delete|. When deleting a group, autocommands contained in
+/// NOTE: behavior differs from |:augroup-delete|. When deleting a group, autocommands contained in
 /// this group will also be deleted and cleared. This group will no longer exist.
 /// @param name String The name of the group.
-/// @see |autocommand-groups|
+/// @see |autocmd-groups|
 void nvim_del_augroup_by_name(String name, Error *err)
   FUNC_API_SINCE(9)
 {
@@ -735,6 +772,8 @@ void nvim_del_augroup_by_name(String name, Error *err)
 ///             {pattern}.
 ///             - modeline (bool) optional: defaults to true. Process the
 ///             modeline after the autocommands |<nomodeline>|.
+///             - data (any): arbitrary data to send to the autocommand callback. See
+///             |nvim_create_autocmd()| for details.
 /// @see |:doautocmd|
 void nvim_exec_autocmds(Object event, Dict(exec_autocmds) *opts, Error *err)
   FUNC_API_SINCE(9)
@@ -743,13 +782,12 @@ void nvim_exec_autocmds(Object event, Dict(exec_autocmds) *opts, Error *err)
   bool modeline = true;
 
   buf_T *buf = curbuf;
-  bool set_buf = false;
-
-  char_u *pattern = NULL;
-  bool set_pattern = false;
 
+  Array patterns = ARRAY_DICT_INIT;
   Array event_array = ARRAY_DICT_INIT;
 
+  Object *data = NULL;
+
   if (!unpack_string_or_array(&event_array, &event, "event", true, err)) {
     goto cleanup;
   }
@@ -787,35 +825,35 @@ void nvim_exec_autocmds(Object event, Dict(exec_autocmds) *opts, Error *err)
     }
 
     buf = find_buffer_by_handle((Buffer)buf_obj.data.integer, err);
-    set_buf = true;
 
     if (ERROR_SET(err)) {
       goto cleanup;
     }
   }
 
-  if (opts->pattern.type != kObjectTypeNil) {
-    if (opts->pattern.type != kObjectTypeString) {
-      api_set_error(err, kErrorTypeValidation, "'pattern' must be a string");
-      goto cleanup;
-    }
+  if (!get_patterns_from_pattern_or_buf(&patterns, opts->pattern, opts->buffer, err)) {
+    goto cleanup;
+  }
 
-    pattern = vim_strsave((char_u *)opts->pattern.data.string.data);
-    set_pattern = true;
+  if (patterns.size == 0) {
+    ADD(patterns, STRING_OBJ(STATIC_CSTR_TO_STRING("")));
+  }
+
+  if (opts->data.type != kObjectTypeNil) {
+    data = &opts->data;
   }
 
   modeline = api_object_to_bool(opts->modeline, "modeline", true, err);
 
-  if (set_pattern && set_buf) {
-    api_set_error(err, kErrorTypeValidation, "must not set 'buffer' and 'pattern'");
-    goto cleanup;
-  }
-
   bool did_aucmd = false;
   FOREACH_ITEM(event_array, event_str, {
     GET_ONE_EVENT(event_nr, event_str, cleanup)
 
-    did_aucmd |= apply_autocmds_group(event_nr, pattern, NULL, true, au_group, buf, NULL);
+    FOREACH_ITEM(patterns, pat, {
+      char *fname = opts->buffer.type == kObjectTypeNil ? pat.data.string.data : NULL;
+      did_aucmd |=
+        apply_autocmds_group(event_nr, fname, NULL, true, au_group, buf, NULL, data);
+    })
   })
 
   if (did_aucmd && modeline) {
@@ -824,13 +862,13 @@ void nvim_exec_autocmds(Object event, Dict(exec_autocmds) *opts, Error *err)
 
 cleanup:
   api_free_array(event_array);
-  XFREE_CLEAR(pattern);
+  api_free_array(patterns);
 }
 
 static bool check_autocmd_string_array(Array arr, char *k, Error *err)
 {
-  for (size_t i = 0; i < arr.size; i++) {
-    if (arr.items[i].type != kObjectTypeString) {
+  FOREACH_ITEM(arr, entry, {
+    if (entry.type != kObjectTypeString) {
       api_set_error(err,
                     kErrorTypeValidation,
                     "All entries in '%s' must be strings",
@@ -839,25 +877,25 @@ static bool check_autocmd_string_array(Array arr, char *k, Error *err)
     }
 
     // Disallow newlines in the middle of the line.
-    const String l = arr.items[i].data.string;
+    const String l = entry.data.string;
     if (memchr(l.data, NL, l.size)) {
       api_set_error(err, kErrorTypeValidation,
                     "String cannot contain newlines");
       return false;
     }
-  }
+  })
   return true;
 }
 
 static bool unpack_string_or_array(Array *array, Object *v, char *k, bool required, Error *err)
 {
   if (v->type == kObjectTypeString) {
-    ADD(*array, copy_object(*v));
+    ADD(*array, copy_object(*v, NULL));
   } else if (v->type == kObjectTypeArray) {
     if (!check_autocmd_string_array(v->data.array, k, err)) {
       return false;
     }
-    *array = copy_array(v->data.array);
+    *array = copy_array(v->data.array, NULL);
   } else {
     if (required) {
       api_set_error(err,
@@ -908,7 +946,7 @@ static int get_augroup_from_object(Object group, Error *err)
 static bool get_patterns_from_pattern_or_buf(Array *patterns, Object pattern, Object buffer,
                                              Error *err)
 {
-  const char_u pattern_buflocal[BUFLOCAL_PAT_LEN];
+  const char pattern_buflocal[BUFLOCAL_PAT_LEN];
 
   if (pattern.type != kObjectTypeNil && buffer.type != kObjectTypeNil) {
     api_set_error(err, kErrorTypeValidation,
@@ -918,7 +956,7 @@ static bool get_patterns_from_pattern_or_buf(Array *patterns, Object pattern, Ob
     Object *v = &pattern;
 
     if (v->type == kObjectTypeString) {
-      char_u *pat = (char_u *)v->data.string.data;
+      char *pat = v->data.string.data;
       size_t patlen = aucmd_pattern_length(pat);
       while (patlen) {
         ADD(*patterns, STRING_OBJ(cbuf_to_string((char *)pat, patlen)));
@@ -932,8 +970,8 @@ static bool get_patterns_from_pattern_or_buf(Array *patterns, Object pattern, Ob
       }
 
       Array array = v->data.array;
-      for (size_t i = 0; i < array.size; i++) {
-        char_u *pat = (char_u *)array.items[i].data.string.data;
+      FOREACH_ITEM(array, entry, {
+        char *pat = entry.data.string.data;
         size_t patlen = aucmd_pattern_length(pat);
         while (patlen) {
           ADD(*patterns, STRING_OBJ(cbuf_to_string((char *)pat, patlen)));
@@ -941,15 +979,15 @@ static bool get_patterns_from_pattern_or_buf(Array *patterns, Object pattern, Ob
           pat = aucmd_next_pattern(pat, patlen);
           patlen = aucmd_pattern_length(pat);
         }
-      }
+      })
     } else {
       api_set_error(err,
                     kErrorTypeValidation,
-                    "'pattern' must be a string");
+                    "'pattern' must be a string or table");
       return false;
     }
   } else if (buffer.type != kObjectTypeNil) {
-    if (buffer.type != kObjectTypeInteger) {
+    if (buffer.type != kObjectTypeInteger && buffer.type != kObjectTypeBuffer) {
       api_set_error(err,
                     kErrorTypeValidation,
                     "'buffer' must be an integer");
@@ -968,9 +1006,9 @@ static bool get_patterns_from_pattern_or_buf(Array *patterns, Object pattern, Ob
   return true;
 }
 
-static bool clear_autocmd(event_T event, char_u *pat, int au_group, Error *err)
+static bool clear_autocmd(event_T event, char *pat, int au_group, Error *err)
 {
-  if (do_autocmd_event(event, pat, false, false, (char_u *)"", true, au_group) == FAIL) {
+  if (do_autocmd_event(event, pat, false, false, "", true, au_group) == FAIL) {
     api_set_error(err, kErrorTypeException, "Failed to clear autocmd");
     return false;
   }
